import pytest
import pdb
import re

# Auxiliary modules
import numpy as np
from copy import deepcopy
import sys

# Module to be tested
sys.path.append('..')
import heppy

# Setup: create a histogram
areas = np.array([4.3, 9.7, 12.1])
binedges = np.array([-7., 0., 5., 50.])
uncorr_variations = {
    'stat_up' : areas + np.sqrt(areas), # approx [6.37, 12.81, 15.58]
    'stat_down' : areas - np.sqrt(areas) # approx [2.23, 6.59, 8.62]
}
corr_variations = {
    'foo_up' : areas + np.ones_like(areas),
    'foo_down' : areas - np.ones_like(areas),
    'single' : areas + np.array([-4., 8.5, 3.0])
}
_h = heppy.histogram(binedges, areas, areas=True, uncorr_variations=uncorr_variations, corr_variations=corr_variations)
_h.name = 'fooble'



# Tests
# TODO: check that illegal options in rebinning methods trigger exception!

def test_heights():
    np.testing.assert_array_almost_equal(_h.heights, np.array([0.6142857142857142, 1.94, 0.2688888888888889]))

def test_find_bin():
    # Check cases where x lies within the histogram x-boundaries:
    assert _h.bin_index(-7.0) == 0
    assert _h.bin_index(-5.0) == 0
    assert _h.bin_index(0.) == 1
    assert _h.bin_index(0.5) == 1
    assert _h.bin_index(5.3) == 2
    # Check cases where x lies outside of the histogram x-boundaries:
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index(-7.1)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = -7.1, which is outside of histogram x-boundaries [-7.0, 50.0)'
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index(60.)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = 60.0, which is outside of histogram x-boundaries [-7.0, 50.0)'
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index(50.) # IMPORTANT: upper bin edge is NOT part of bin!
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = 50.0, which is outside of histogram x-boundaries [-7.0, 50.0)'

def test_integral():
    np.testing.assert_approx_equal(_h.integral(), 26.1)

def test_rebin():
    h = deepcopy(_h)
    h.rebin([-7., 5., 50.])
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7, 5., 50.]))
    np.testing.assert_array_almost_equal(h.areas, np.array([14., 12.1]))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], h.areas + np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], h.areas - np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_up'], np.array([16., 13.1]))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_down'], np.array([12., 11.1]))
    np.testing.assert_array_almost_equal(h.corr_variations['single'], np.array([18.5, 15.1]))
    np.testing.assert_approx_equal(h.integral(), 26.1)

def test_lossy_rebin():
    """
    Rebinning less than the complete bin range of the histogram.
    Information falling outside of the new outer bin edges is dropped.
    """
    # Cutting off from the low end
    h = deepcopy(_h)
    h.rebin([5., 50.])
    np.testing.assert_array_almost_equal(h.binedges, np.array([5., 50.]))
    np.testing.assert_array_almost_equal(h.areas, np.array([12.1]))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], h.areas + np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], h.areas - np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_up'], np.array([13.1]))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_down'], np.array([11.1]))
    np.testing.assert_array_almost_equal(h.corr_variations['single'], np.array([15.1]))
    np.testing.assert_approx_equal(h.integral(), 12.1)
    # Cutting off from the high end
    h = deepcopy(_h)
    h.rebin([-7., 5.])
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7., 5.]))
    np.testing.assert_array_almost_equal(h.areas, np.array([14.]))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], h.areas + np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], h.areas - np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_up'], np.array([16.]))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_down'], np.array([12.]))
    np.testing.assert_array_almost_equal(h.corr_variations['single'], np.array([18.5]))
    np.testing.assert_approx_equal(h.integral(), 14.)


def test_merge_bins():
    h = deepcopy(_h)
    h.merge_bins(0., 50.)
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7, 0., 50.]))
    np.testing.assert_array_almost_equal(h.areas, np.array([4.3, 21.8]))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], h.areas + np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], h.areas - np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_up'], np.array([5.3, 23.8]))
    np.testing.assert_array_almost_equal(h.corr_variations['foo_down'], np.array([3.3, 19.8]))
    np.testing.assert_array_almost_equal(h.corr_variations['single'], np.array([0.3, 33.3]))
    np.testing.assert_approx_equal(h.integral(), 26.1)

def test_squash_highest_bin():
    h = deepcopy(_h)
    h.squash_highest_bin(0., 25.)
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7, 0., 25.]))
    np.testing.assert_array_almost_equal(h.areas, np.array([4.3, 21.8]))
    np.testing.assert_array_almost_equal(h.heights, np.array([0.6142857142857142, 0.872]))
    np.testing.assert_approx_equal(h.integral(), 26.1)

def test_net_variations_uncorr_only():
    h = deepcopy(_h)
    h.corr_variations = {}
    upper, lower = h.net_variations()
    np.testing.assert_array_almost_equal(lower, h.areas - np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(upper, h.areas + np.sqrt(h.areas))
    upper_shift, lower_shift = h.net_variations(subtract_nominal=True)
    np.testing.assert_array_almost_equal(lower_shift, -np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(upper_shift, np.sqrt(h.areas))
    upper_rel, lower_rel = h.net_variations(relative=True)
    np.testing.assert_array_almost_equal(lower_rel, (h.areas - np.sqrt(h.areas)) / h.areas)
    np.testing.assert_array_almost_equal(upper_rel, (h.areas + np.sqrt(h.areas)) / h.areas)
    upper_rel_shift, lower_rel_shift = h.net_variations(subtract_nominal=True, relative=True)
    np.testing.assert_array_almost_equal(lower_rel_shift, -1./np.sqrt(h.areas))
    np.testing.assert_array_almost_equal(upper_rel_shift, 1./np.sqrt(h.areas))

def test_net_variations_corr_only():
    h = deepcopy(_h)
    h.uncorr_variations = {}
    # TODO!
    # (lower, upper) = h.net_variations()
    # np.testing.assert_array_almost_equal(lower, h.areas - np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper, h.areas + np.sqrt(h.areas))
    # (lower_shift, upper_shift) = h.net_variations(subtract_nominal=True)
    # np.testing.assert_array_almost_equal(lower_shift, -np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper_shift, np.sqrt(h.areas))
    # (lower_rel, upper_rel) = h.net_variations(relative=True)
    # np.testing.assert_array_almost_equal(lower_rel, (h.areas - np.sqrt(h.areas)) / h.areas)
    # np.testing.assert_array_almost_equal(upper_rel, (h.areas + np.sqrt(h.areas)) / h.areas)
    # (lower_rel_shift, upper_rel_shift) = h.net_variations(subtract_nominal=True, relative=True)
    # np.testing.assert_array_almost_equal(lower_rel_shift, -1./np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper_rel_shift, 1./np.sqrt(h.areas))

def test_net_variations_selective():
    h = deepcopy(_h)
    h.corr_variations = {}
    upper, lower = h.net_variations()
    _upper, _lower = _h.net_variations(['stat_up', 'stat_down'])
    np.testing.assert_array_almost_equal(lower, _lower)
    np.testing.assert_array_almost_equal(upper, _upper)

def test_net_variations():
    pass
    # TODO!
    # (lower, upper) = h.net_variations()
    # np.testing.assert_array_almost_equal(lower, h.areas - np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper, h.areas + np.sqrt(h.areas))
    # (lower_shift, upper_shift) = h.net_variations(subtract_nominal=True)
    # np.testing.assert_array_almost_equal(lower_shift, -np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper_shift, np.sqrt(h.areas))
    # (lower_rel, upper_rel) = h.net_variations(relative=True)
    # np.testing.assert_array_almost_equal(lower_rel, (h.areas - np.sqrt(h.areas)) / h.areas)
    # np.testing.assert_array_almost_equal(upper_rel, (h.areas + np.sqrt(h.areas)) / h.areas)
    # (lower_rel_shift, upper_rel_shift) = h.net_variations(subtract_nominal=True, relative=True)
    # np.testing.assert_array_almost_equal(lower_rel_shift, -1./np.sqrt(h.areas))
    # np.testing.assert_array_almost_equal(upper_rel_shift, 1./np.sqrt(h.areas))

def test_errorbars_uncorr_only():
    # TODO: change to contain ALL variations, also correlated ones
    h = deepcopy(_h)
    h.corr_variations = {}
    upper, lower = h.errorbars()
    np.testing.assert_array_almost_equal(lower, np.sqrt(h.areas)/h.binwidths)
    np.testing.assert_array_almost_equal(upper, np.sqrt(h.areas)/h.binwidths)

def test_add():
    h = _h + _h
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(h.areas, 2*areas)
    np.testing.assert_array_almost_equal(h.heights, 2.* np.array([0.6142857142857142, 1.94, 0.2688888888888889]))
    h.corr_variations = {} # TODO: delete me!
    upper_errorbar, lower_errorbar = h.errorbars()
    np.testing.assert_array_almost_equal(lower_errorbar, np.sqrt(h.areas)/h.binwidths)
    np.testing.assert_array_almost_equal(upper_errorbar, np.sqrt(h.areas)/h.binwidths)

def test_add_with_zero_bin_area():
    g = deepcopy(_h)
    g.areas = np.array([2., 1., 0.])
    g.uncorr_variations = {
        'stat_up' : np.array([2., 1., 0.]) + np.sqrt(np.array([2., 1., 0.])),
        'stat_down' : np.array([2., 1., 0.]) - np.sqrt(np.array([2., 1., 0.])),
    }
    h = g + _h
    h.corr_variations = {} # TODO: delete me!
    upper_errorbar, lower_errorbar = h.errorbars()
    np.testing.assert_array_almost_equal(lower_errorbar, np.sqrt(h.areas)/h.binwidths)
    np.testing.assert_array_almost_equal(upper_errorbar, np.sqrt(h.areas)/h.binwidths)

def test_multiply():
    h = _h * 3.5
    h.corr_variations = {} # TODO: delete me!
    np.testing.assert_array_almost_equal(h.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(h.areas, 3.5*areas)
    np.testing.assert_array_almost_equal(h.heights, 3.5* np.array([0.6142857142857142, 1.94, 0.2688888888888889]))
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], 3.5*_h.uncorr_variations['stat_up'])
    np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], 3.5*_h.uncorr_variations['stat_down'])
    upper, lower = h.net_variations('stat_up')
    _upper, _lower = _h.net_variations('stat_up')
    np.testing.assert_array_almost_equal(lower, 3.5*_lower)
    np.testing.assert_array_almost_equal(upper, 3.5*_upper)
    upper_errorbar, lower_errorbar = h.errorbars()
    _upper_errorbar, _lower_errorbar = _h.errorbars(['stat_up', 'stat_down'])
    np.testing.assert_array_almost_equal(lower_errorbar, 3.5*_lower_errorbar)
    np.testing.assert_array_almost_equal(upper_errorbar, 3.5*_upper_errorbar)

def test_divide_scalar():
    h = deepcopy(_h)
    h.corr_variations = {} # TODO: delete me!
    r = h / 3.0
    np.testing.assert_array_almost_equal(r.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(r.areas, areas/3.0)

def test_histdiv_scalar():
    '''
    This is not typically an interesting use case, but the unit test may be useful for pinpointing issues
    '''
    h = deepcopy(_h)
    h.corr_variations = {} # TODO: delete me!
    r = heppy.histdiv(h, 3.0)
    np.testing.assert_array_almost_equal(r.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(r.heights, areas/3.0) # NOTE: heights, not areas, due to non-division by binwidths

def test_divide():
    h = deepcopy(_h)
    h.corr_variations = {} # TODO: delete me!
    g = deepcopy(_h)
    g.corr_variations = {} # TODO: delete me!
    g.areas = np.array([2., 1., 0.])
    g.uncorr_variations = {
        'stat_up' : np.array([2., 1., 0.]) + np.sqrt(np.array([2., 1., 0.])),
        'stat_down' : np.array([2., 1., 0.]) - np.sqrt(np.array([2., 1., 0.])),
    }
    r = g / h
    np.testing.assert_array_almost_equal(r.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(r.areas, np.array([0.46511628, 0.10309278, 0.]))
    # np.testing.assert_array_almost_equal(h.heights, 3.5* np.array([0.6142857142857142, 1.94, 0.2688888888888889]))
    # np.testing.assert_array_almost_equal(h.uncorr_variations['stat_up'], 3.5*_h.uncorr_variations['stat_up'])
    # np.testing.assert_array_almost_equal(h.uncorr_variations['stat_down'], 3.5*_h.uncorr_variations['stat_down'])
    # upper, lower = h.net_variations('stat_up')
    # _upper, _lower = _h.net_variations('stat_up')
    # np.testing.assert_array_almost_equal(lower, 3.5*_lower)
    # np.testing.assert_array_almost_equal(upper, 3.5*_upper)
    # upper_errorbar, lower_errorbar = h.errorbars()
    # _upper_errorbar, _lower_errorbar = _h.errorbars(['stat_up', 'stat_down'])
    # np.testing.assert_array_almost_equal(lower_errorbar, 3.5*_lower_errorbar)
    # np.testing.assert_array_almost_equal(upper_errorbar, 3.5*_upper_errorbar)

def test_cumulative():
    h = deepcopy(_h)
    # Intergrating/summing from the left:
    lcum = h.cumulative()
    np.testing.assert_array_almost_equal(lcum.binedges, h.binedges)
    np.testing.assert_array_almost_equal(lcum.heights, np.array([4.3, 4.3+9.7, 4.3+9.7+12.1]))
    # Intergrating/summing from the right:
    rcum = h.cumulative(side='right')
    np.testing.assert_array_almost_equal(rcum.binedges, h.binedges)
    np.testing.assert_array_almost_equal(rcum.heights, np.array([4.3+9.7+12.1, 9.7+12.1, 12.1]))

def test_histdiv():
    '''
    This is not typically an interesting use case, but the unit test may be useful for pinpointing issues
    '''
    h = deepcopy(_h)
    h.corr_variations = {} # TODO: delete me!
    g = deepcopy(_h)
    g.corr_variations = {} # TODO: delete me!
    g.areas = np.array([2., 1., 0.])
    g.uncorr_variations = {
        'stat_up' : np.array([2., 1., 0.]) + np.sqrt(np.array([2., 1., 0.])),
        'stat_down' : np.array([2., 1., 0.]) - np.sqrt(np.array([2., 1., 0.])),
    }
    r = heppy.histdiv(g, h)
    np.testing.assert_array_almost_equal(r.binedges, np.array([-7, 0., 5., 50.]))
    np.testing.assert_array_almost_equal(r.heights, np.array([0.46511628, 0.10309278, 0.])) # NOTE: heights, not areas, due to non-division by binwidths

def test_to_root():
    """Tests converting the histogram to a dictionary of ROOT.TH1D's mapped by variation key."""
    th1s = _h.to_root()
    assert isinstance(th1s, dict)
    assert len(th1s) == 6
    assert sorted(list(th1s.keys())) == ['fooble_foo_down', 'fooble_foo_up', 'fooble_nominal', 'fooble_single', 'fooble_stat_down', 'fooble_stat_up']
    np.testing.assert_approx_equal(th1s['fooble_nominal'].GetBinContent(2), 9.7)
    np.testing.assert_approx_equal(th1s['fooble_foo_down'].GetBinContent(3), 11.1)

def test_to_root_file():
    """Tests writing the histograms to a bunch of ROOT.TH1D's inside a ROOT file."""
    _h.to_root_file('output.root', recreate=True)
    _g = deepcopy(_h)
    _g.name = 'barbaz'
    # Add to same file
    _g.to_root_file('output.root', nominal_label='default', key_form='{name}__{variation}')
    # TODO add some tests



#!/usr/bin/env python3

import numpy as np
import heppy as hp
import os


def readtrex(trexfile, unfolded_format=False):
    """Reads Heppy histogram from TRExFitter YAML format.

    Requires PyYAML.

    IMPORTANT: the uncertainties are treated as correlated uncertainties, since
    their actual correlations cannot be inferred from the histogram. So treat
    them with care: they're fine for plotting and looking at individual bins,
    but do not rebin the histogram or do statistical testing involving its shape
    and assume that you will find correct results.

    :param trexfile: path to input file.
    :type trexfile: ``str``
    :param unfolded_format: set this to True if reading the YAML format that
    TRExFitter uses for unfolded cross sections.
    :type unfolded_format: ``bool``
    """
    import yaml
    if not unfolded_format:
        raise NotImplementedError('Reading TRExFitter YAML has only been '
            'implemented for its format used for unfolded cross sections so '
            'far.')
    with open(trexfile, 'r') as infile:
        bin_properties = yaml.safe_load(infile)
        binedges = [bin_properties[0]['range'][0]] # lower edge of first bin
        nominals = []
        ups = []
        downs = []
        for bin_property in bin_properties:
            binedges.append(bin_property['range'][1]) # append upper bin edge
            nominal = bin_property['mean']
            upshift = bin_property['uncertaintyUp'] # signed!
            downshift = bin_property['uncertaintyDown'] # signed!
            nominals.append(nominal)
            ups.append(nominal+upshift)
            downs.append(nominal+downshift)
        name = os.path.splitext(trexfile)[0]
        provenance = os.path.expandvars(os.path.abspath(trexfile))
        return hp.histogram1d(np.array(binedges), np.array(nominals),
            areas=True, name=name, corr_variations={
            'TRExFitter_unfolding__up' : np.array(ups),
            'TRExFitter_unfolding__down' : np.array(downs)},
            attributes={'provenance' : provenance})
    raise ValueError('Could not successfully read histogram from TRExFitter '
        f'YAML file {trexfile}.')



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from .histogram import histogram2d
import heppy.panel
import heppy.plot



class TextFormatter:
    '''
    Predefined functions to format bin content text printed on heatmap. The user can alternatively write their own such functions.

    Contents will be printed with a precision of (up to) three significant digits. If you want to set a different precision, you can
    create your own adapted formatting function as the following example illustrates:

    .. code-block:: python

        import functools
        # set the number of significant digits (e.g. to 4)
        formatter = functools.partial(TextFormatter.nominal, significants=4)
        # or set the fixed absolute precision (e.g. to 3 digits after the decimal point)
        formatter = functools.partial(TextFormatter.nominal, fixedprec=3)
        # then use as: heppy.make_heatmap(..., text_format=formatter, ...)

    Here :code:`significants` represents the maximum number of significant digits considered (default: 3), while :code:`fixedprec`
    represents the fixed absolute precision considered, e.g. :code:`1` for a precision of 0.1 or :code:`-1` for a precision of 10
    (default: :code:`None`). If :code:`fixedprec` is given, :code:`significants` is ignored.

    Hint: when writing a custom formatting function, any unnecessary keyword arguments can be absorbed into
    a :code:`**kwargs` catch-all parameter to keep the function signature shorter and tidier.
    '''

    @staticmethod
    def _to_str(val, significants=3, fixedprec=None, suppress_zero=True):
        '''
        Auxiliary function to get a string representation of the value with a given
        number of significant digits or with a given fixed precision

        Is suppress_zero is True, texts that round to a number mathematically equivalent to zero
        are not printed
        '''
        if fixedprec is not None:
            fixedprec = int(fixedprec)
            if fixedprec >= 0:
                string = np.format_float_positional(val, precision=fixedprec, trim='-')
                return '' if suppress_zero and set(string) <= {'-', '.', '0'} else string
            val = round(val, fixedprec)
            return '{:g}'.format(val)
            #return '{:.<digits>f}'.replace('<digits>', str(max(0, fixedprec))).format(val)
        string = np.format_float_positional(float('{:.<sig>g}'.replace('<sig>', str(significants)).format(val)), precision=20, trim='-')
        return '' if suppress_zero and set(string) <= {'-', '.', '0'} else string



    @staticmethod
    def _pm_uncertainty(up, down):
        '''
        Return "+/- up" if up == down, else return "^{+u}_{-d}"
        '''
        return r'\pm {up}'.format(up=up) if up == down else r'^{{+{up}}}_{{-{down}}}'.format(up=up, down=down)



    @staticmethod
    def nominal(significants=3, fixedprec=None, nominal=None, **ignore):
        """Returns LaTeX string of nominal value."""
        return TextFormatter._to_str(nominal, significants=significants, fixedprec=fixedprec)



    @staticmethod
    def brief(significants=3, fixedprec=None, nominal=None, uncert_up=None, uncert_down=None, **ignore):
        r"""Returns LaTeX string of nominal value and total uncertainty.

        The format is :math:`\mathrm{nominal} \pm \sigma`,
        where :math:`\sigma` is the uncertainty from the uncorrelated variations and
        the correlated variations.
        Asymmetric uncertainties are supported and will be shown as
        :math:`^{\sigma^{\mathrm{up}}}_{\sigma^{\mathrm{down}}}`.
        """
        n = TextFormatter._to_str(nominal, significants=significants, fixedprec=fixedprec)
        u = TextFormatter._to_str(uncert_up, significants=significants, fixedprec=fixedprec)
        d = TextFormatter._to_str(uncert_down, significants=significants, fixedprec=fixedprec)
        # If _all_ of `n`, `u`, `d` rounded to zero and were suppressed, return empty string.
        if not any([n, u, d]):
            return ''
        # If _some but not all_ of `n`, `u`, `d` rounded to zero and were suppressed,
        # remake them all without suppression. Otherwise we get awkward texts like "± 0.1" or
        # "5.0 ±" with missing numbers.
        if any([n, u, d]) and not all([n, u, d]):
            n = TextFormatter._to_str(nominal, significants=significants, fixedprec=fixedprec, suppress_zero=False)
            u = TextFormatter._to_str(uncert_up, significants=significants, fixedprec=fixedprec, suppress_zero=False)
            d = TextFormatter._to_str(uncert_down, significants=significants, fixedprec=fixedprec, suppress_zero=False)
        return r'${n}{pm_uncert}$'.format(n=n, pm_uncert=TextFormatter._pm_uncertainty(u, d))



    @staticmethod
    def statsyst(significants=3, fixedprec=None, nominal=None, stat_up=None, stat_down=None, syst_up=None, syst_down=None, **ignore):
        r"""Returns LaTeX string of nominal value and statistical and systematic uncertainty.

        The format is :math:`\mathrm{nominal} \pm \sigma_{\mathrm{stat}} \pm \sigma_{\mathrm{syst}}`,
        where :math:`\sigma_{\mathrm{stat}}` is the uncertainty from the uncorrelated variations and
        :math:`\sigma_{\mathrm{syst}}` is the uncertainty from the correlated variations.
        Asymmetric uncertainties are supported and will be shown as
        :math:`^{\sigma_{\mathrm{syst}}^{\mathrm{up}}}_{\sigma_{\mathrm{syst}}^{\mathrm{down}}}` etc.
        """
        n = TextFormatter._to_str(nominal, significants=significants, fixedprec=fixedprec)
        stat_u = TextFormatter._to_str(stat_up, significants=significants, fixedprec=fixedprec)
        stat_d = TextFormatter._to_str(stat_down, significants=significants, fixedprec=fixedprec)
        syst_u = TextFormatter._to_str(syst_up, significants=significants, fixedprec=fixedprec)
        syst_d = TextFormatter._to_str(syst_down, significants=significants, fixedprec=fixedprec)
        if all([s == '' for s in [n, stat_u, stat_d, syst_u, syst_d]]): return ''
        return r'${n}{pm_stat_uncert}{{}}{pm_syst_uncert}$'.format(n=n, pm_stat_uncert=TextFormatter._pm_uncertainty(stat_u, stat_d), pm_syst_uncert=TextFormatter._pm_uncertainty(syst_u, syst_d))





def _make_text(text_format, input_dict):
    '''
    Create a string from the values in input dictionary according to the specified formatting string or function

    :param text_format: string template or function returning the text to be printed inside each bin
    :type text_format: `str` or `function`

    :param input_dict: dictionary containing the following keys: `'nominal'`, `'uncert_up'`, `'uncert_down'`, `'stat_up'`, `'stat_down'`, `'syst_up'`, `'syst_down'`
    :type input_dict: `dict`
    '''
    if isinstance(text_format, str):
        return text_format.format(**input_dict)
    # Assume it's a function and call it with the contents of input_dict as keyword arguments
    return text_format(**input_dict)



def _map_to_equispaced(array):
    '''
    Map each value in the array to an integer-and-a-half value such that the minimum value gets
    mapped to 0.5 and the highest value to len(array) - 0.5, and all other values fall in between
    in a sorted manner.
    '''
    # NB: np.unique returns the _sorted_ unique values
    translate = { val : i + 0.5 for i, val in enumerate(np.unique(array)) }
    return np.array( [translate[val] for val in array] )



def make_heatmap(histogram, areas=False, title='', figsize=(8, 5), colorbar=True, text_format=TextFormatter.brief, text_precision=3, text_autocolor=True, black_text_below=None, monowidth=False, write='', xmax=None, **kwargs):
    '''
    Make a heatmap plot of a two-dimensional histogram

    :param histogram: two-dimensional histogram to visualise
    :type histogram: :py:class:`heppy.histogram2d`
    :param title: plot title
    :type title: :code:`str`
    :param figsize: figure size
    :type figsize: :code:`tuple` of :code:`float`
    :param colorbar: show color bar or not
    :type colorbar: ``bool``
    :param text_format: string template or function returning the text to be printed inside each bin. The following format keys (if string) or keyword arguments \
    (if function) will be provided: {nominal}, {uncert_up} and {uncert_down} for total uncertainty, {stat_up} and {stat_down} for statistical uncertainty, \
    {syst_up} and {syst_down} for systematic uncertainty. All uncertainties are given as non-negative numbers. The class :py:class:`heppy.TextFormatter` provides \
    a set of useful and somewhat adaptable predefined formatter functions.
    :type text_format: :code:`str`, :code:`function` or :code:`None`
    :param monowidth: if True, all bins are shown as equally wide/high, with the bin edges written in the label.
    :type monowidth: :code:`bool`
    :param write: may be changed to a filename, which will result in the figure being rendered and saved to disk
    :type write: :code:`str`
    :param **kwargs: keyword arguments that get passed on to :code:`plt.hist2d`

    The following keys in :code:`histogram.plot_attributes` can be used to set axis labels:

    - :code:`"xlabel"` – x-axis label
    - :code:`"ylabel"` – y-axis label
    '''
    if not isinstance(histogram, histogram2d):
        raise TypeError('Heatmaps can only be made for a 2D histogram (heppy.histogram2d), got instance of type {type}'.format(type=type(histogram)))

    fig, axes = plt.subplots(1, figsize=figsize)

    midx, midy, heights = histogram.points()


    # Visualised bin edges and midpoints
    vis_binedges = histogram.binedges
    vis_midx = midx
    vis_midy = midy
    # If monowidth visualisation is desired, the actual bin edges and midpoints are mapped to equidistant
    # visualised ones:
    if monowidth:
        # Bin edges at integers from zero:
        vis_binedges = (np.array([i for i, _ in enumerate(histogram.binedges[0])]), np.array([i for i, _ in enumerate(histogram.binedges[1])]))
        # Bin midpoints at integers from zero PLUS 0.5 (to make them midpoints w.r.t. the edges):
        vis_midx = _map_to_equispaced(midx)
        vis_midy = _map_to_equispaced(midy)


    # Set default plot options and override them with the ones the user provided
    plot_options = {'norm' : LogNorm(), 'cmap' : 'Blues', **kwargs}

    # midx and midy are used here as "dummy" fill values at the bin centres.
    # The histogram is filled "with" these values with the actual bin height (or optionally area) as "weight"
    nominals = np.ravel(histogram.areas) if areas else heights
    uncertups, uncertdowns = histogram.net_variations(variations='all', subtract_nominal=True, relative=False)
    uncertups = np.ravel(uncertups) if areas else np.ravel(uncertups/histogram.binsizes)
    uncertdowns = np.abs(np.ravel(uncertdowns)) if areas else np.abs(np.ravel(uncertdowns/histogram.binsizes))
    systups, systdowns = histogram.net_variations(variations=list(histogram.corr_variations.keys()), subtract_nominal=True, relative=False)
    systups = np.ravel(systups) if areas else np.ravel(systups/histogram.binsizes)
    systdowns = np.abs(np.ravel(systdowns)) if areas else np.abs(np.ravel(systdowns/histogram.binsizes))
    statups, statdowns = histogram.net_variations(variations=list(histogram.uncorr_variations.keys()), subtract_nominal=True, relative=False)
    statups = np.ravel(statups) if areas else np.ravel(statups/histogram.binsizes)
    statdowns = np.abs(np.ravel(statdowns)) if areas else np.abs(np.ravel(statdowns/histogram.binsizes))

    h2d = plt.hist2d(vis_midx, vis_midy, weights=nominals, bins=vis_binedges, **plot_options)
    if colorbar:
        fig.colorbar(h2d[3])

    max_value_of_heatmap = np.max(nominals)
    if text_format is not None:
        # Write bin contents as text onto the plot
        for x, y, nominal, uncert_up, uncert_down, syst_up, syst_down, stat_up, stat_down in zip(vis_midx, vis_midy, nominals, uncertups, uncertdowns, systups, systdowns, statups, statdowns):
            values = {
                'nominal' : nominal,
                'uncert_up' : uncert_up,
                'uncert_down' : uncert_down,
                'syst_up' : syst_up,
                'syst_down' : syst_down,
                'stat_up' : stat_up,
                'stat_down' : stat_down,
            }
            text = _make_text(text_format, values)
            if black_text_below is None:
                color = 'black' if (nominal / max_value_of_heatmap < 0.03) else 'white'
            else:
                color = 'black' if (nominal < black_text_below) else 'white'
            if not text_autocolor:
                color = 'black'
            axes.text(x, y, text, verticalalignment='center', horizontalalignment='center', color=color, fontsize=8)

    # Set proper tick labels for monowidth plots:
    if monowidth:
        axes.set_xticks(vis_binedges[0])
        axes.set_yticks(vis_binedges[1])
        # Nicely format the actual binedges: drop any trailing decimal points and zeros
        # if all have integer values by converting the values to ints:
        actual_binedges_x = histogram.binedges[0].astype(int) if np.all(np.equal(np.mod(histogram.binedges[0], 1), 0)) else np.around(histogram.binedges[0], 5)
        actual_binedges_y = histogram.binedges[1].astype(int) if np.all(np.equal(np.mod(histogram.binedges[1], 1), 0)) else np.around(histogram.binedges[1], 5)
        axes.set_xticklabels(actual_binedges_x)
        axes.set_yticklabels(actual_binedges_y)

    # Make the plot prettier
    if title:
        axes.set_title(title, x=0., ha='left', size='x-large')
    if 'xlabel' in histogram.plot_attributes:
        axes.set_xlabel(histogram.plot_attributes['xlabel'], ha='right', x=1., size='x-large')
    if 'ylabel' in histogram.plot_attributes:
        axes.set_ylabel(histogram.plot_attributes['ylabel'], ha='right', y=1., size='x-large')
    axes.tick_params(which='both', bottom=True, top=True, left=True, right=True, direction='in')
    fig.subplots_adjust(top=0.9, hspace=0.02)

    # Maybe save it to file
    if write:
        fig.savefig(write)

    return fig, axes
